import logging
import os
import stat
from threading import Thread

from .backends.brewpi_legacy import BrewPiLegacySyncherBackend
from .backends.mock import VirtualBrewPiSyncherBackend
from .backstores.database import DatabaseSyncher
from .forwarders.sse import SSEForwarder

LOGGER = logging.getLogger(__name__)

from circuits import Component, handler
from brewpi_service.controller.events import ControllerRequestCurrentProfile

from brewpi_service.controller.state import ControllerStateManager


class DataSyncherServer(Component):
    """
    Main synching process that spawns hardware backends in threadsand triggers
    backstores on events.
    """
    def init(self):
        self._state_manager = ControllerStateManager().register(self)

        self._backstores = (DatabaseSyncher().register(self))
        self._forwarders = (SSEForwarder().register(self))
        self._backends = (VirtualBrewPiSyncherBackend(self._state_manager).register(self))
        # self._backends = (BrewPiLegacySyncherBackend(self._state_manager).register(self))


    @handler("started")
    def started(self, *args):
        LOGGER.info("Data Syncher running...")



from circuits import Event


class ControllerConnected(Event):
    """
    When a controller connects
    """
    def __init__(self, aController):
        super(ControllerConnected, self).__init__()
        self.controller = aController


class ControllerDisconnected(Event):
    def __init__(self, aController):
        super(ControllerDisconnected, self).__init__()
        self.controller = aController


class ControllerBlockList(Event):
    def __init__(self, aController, aBlockList):
        super(ControllerBlockList, self).__init__()
        self.controller = aController
        self.blocks = aBlockList


class ControllerStateChangeRequest(Event):
    def __init__(self, aController, changes):
        super(ControllerStateChangeRequest, self).__init__()
        self.controller = aController
        self.changes = changes


class ControllerRequestCurrentProfile(Event):
    pass


class ControllerCleanStaleAvailableBlocks(Event):
    """
    Ask persisting backstores to remove all stale blocks
    """
    def __init__(self, aController, time_limit):
        super(ControllerCleanStaleAvailableBlocks, self).__init__()
        self.controller = aController
        self.time_limit = time_limit





from brewpi_service.database import db_session
from brewpi_service.admin import admin, ModelView

from .models import TemperatureSensor, PID


class TemperatureSensorAdminView(ModelView):
    column_hide_backrefs = False
    column_exclude_list = ('type',)
    # column_filters = ('controller',)


admin.add_view(TemperatureSensorAdminView(TemperatureSensor,
                                          db_session))


class PIDAdminView(ModelView):
    column_hide_backrefs = True
    column_exclude_list = ('type',)
    # column_filters = ('controller',)


admin.add_view(PIDAdminView(PID,
                            db_session))



from flask_sse import sse

from brewpi_service import app

app.register_blueprint(sse, url_prefix='/stream')

from sqlalchemy import event

from .controller.models import Controller

from logging import getLogger

LOGGER = getLogger(__name__)



from brewpi_service.admin import admin, ModelView
from brewpi_service.database import db_session

from .models import Controller

# Add basic models to admin
admin.add_view(ModelView(Controller, db_session))



from unittest import TestCase
from unittest.mock import Mock

from ..state import (
    ControllerState,
    ControllerStateManager,
    ControllerStateTransaction
)


class TestControllerState(TestCase):
    def test_simple_creation(self):
        manager = ControllerStateManager()
        state = ControllerState(manager)

    def test_begin_transaction(self):
        manager = ControllerStateManager()
        state = ControllerState(manager)

        transaction = state.begin_transaction()

        self.assertTrue(isinstance(transaction, ControllerStateTransaction))
        self.assertEqual(transaction.get_compiled_changes(), [])


class TestControllerStateTransaction(TestCase):
    def setUp(self):
        self.transaction = ControllerStateTransaction()

    def test_no_block_no_change(self):
        changes = self.transaction.get_compiled_changes()
        self.assertEqual(changes, [])

    def test_one_block_no_change(self):
        block = Mock()
        block.get_dirty_fields = Mock(return_value=[])
        self.transaction.add(block)

        changes = self.transaction.get_compiled_changes()

        block.get_dirty_fields.assert_called_once()
        self.assertEqual(changes, [])

    def test_one_block_one_change(self):
        block = Mock()
        block.changed_field = Mock()
        block.changed_field.get_requested_value = Mock(return_value='new_value')
        block.get_dirty_fields = Mock(return_value=['changed_field'])

        self.transaction.add(block)

        changes = self.transaction.get_compiled_changes()

        self.assertEqual(len(changes), 1)
        self.assertEqual(len(changes[0]), 3)
        self.assertEqual(changes[0][0], block)
        self.assertEqual(changes[0][1], 'changed_field')
        self.assertEqual(changes[0][2], 'new_value')

        block.get_dirty_fields.assert_called_once()
        block.changed_field.get_requested_value.assert_called_once()

    def test_clear(self):
        block = Mock()
        block.get_dirty_fields = Mock(return_value=[])

        self.transaction.add(block)
        changes = self.transaction.get_compiled_changes()

        block.get_dirty_fields.assert_called_once()
        self.assertEqual(changes, [])

        # Now, clear
        self.transaction.clear()
        changes = self.transaction.get_compiled_changes()
        self.assertEqual(changes, [])




import logging

from sqlalchemy import create_engine, Column
from sqlalchemy.orm import scoped_session, sessionmaker, mapper, composite
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.ext.declarative import declarative_base, DeclarativeMeta
from sqlalchemy.exc import IntegrityError

from flask_plugins import emit_event

from brewpi_service import app

LOGGER = logging.getLogger(__name__)

engine = create_engine(app.config['DATABASE_URI'], convert_unicode=True)
db_session = scoped_session(sessionmaker(autocommit=False,
                                         autoflush=False,
                                         bind=engine))


@app.teardown_appcontext
def shutdown_session(exception=None):
    db_session.remove()

from collections import UserList


class ControllerCompositeColumn(object):
    def __init__(self, actual_column, requested_column):
        self.actual_column = actual_column
        self.requested_column = requested_column

    def __composite_values__(self):
        return self.actual_column, self.requested_column

from brewpi_service.controller.state import ControllerDataField
class BaseClsMeta(DeclarativeMeta):
    def __init__(cls, classname, bases, dict_):
        cls._controller_data_fields = []

        for name, attribute in cls.__dict__.items():
            if type(attribute) is ControllerDataField:
                LOGGER.debug("Marking attribute '{0}' as controller state".format(name))
                cls._controller_data_fields.append(name)

        return super(BaseClsMeta, cls).__init__(classname, bases, dict_)


Base = declarative_base(metaclass=BaseClsMeta, mapper=mapper)
Base.query = db_session.query_property()


def load_models():
    from .controller import models # NOQA
    emit_event('load-database-models')


def init_db():
    """
    import all modules here that might define models so that
    they will be registered properly on the metadata.  Otherwise
    you will have to import them first before calling init_db()
    """
    Base.metadata.create_all(bind=engine)


def get_or_create(session,
                  model,
                  create_method='',
                  create_method_kwargs=None,
                  **kwargs):
    """
    Shortcut to get or create an object
    """
    try:
        return session.query(model).filter_by(**kwargs).one(), False
    except NoResultFound:
        kwargs.update(create_method_kwargs or {})
        created = getattr(model, create_method, model)(**kwargs)
        try:
            session.add(created)
            session.commit()
            return created, True
        except IntegrityError as e:
            session.rollback()
            LOGGER.error(e)
            return session.query(model).filter_by(**kwargs).one(), True



from marshmallow_polyfield import PolyField

from brewpi_service import ma

from .models import (
    Controller,
    ControllerBlock,
    ControllerProfile
)

class ControllerBlockDisambiguator:
    class_to_schema = {
    }


def controller_block_schema_serialization_disambiguation(base_object, parent_obj):
    try:
        return ControllerBlockDisambiguator.class_to_schema[base_object.__class__.__name__]()
    except KeyError:
        pass

    raise TypeError("Could not detect type of {0}. "
                    "Did not have a base or a length. "
                    "Are you sure this is a Controller Block?".format(base_object.__class__))


class ControllerSchema(ma.ModelSchema):
    class Meta:
        model = Controller
        fields = ('id', 'connected', 'profile', 'available_blocks', 'name', 'description', 'uri')

    profile = ma.HyperlinkRelated('controllerprofiledetail')
    available_blocks = ma.AbsoluteUrlFor('controlleravailableblocklist', controller_id='<id>')


class ControllerProfileSchema(ma.ModelSchema):
    class Meta:
        model = ControllerProfile
        fields = ('name', 'blocks')

    blocks = ma.List(PolyField(
        serialization_schema_selector=controller_block_schema_serialization_disambiguation,
    ))


class ControllerBlockSchema(ma.ModelSchema):
    class Meta:
        model = ControllerBlock
        fields = ('type', 'updated_at', 'is_static', 'object_id', 'name', 'url')

    type = ma.Function(lambda obj: obj.__class__.__name__)
    url = ma.AbsoluteUrlFor('controllerblockdetail', id='<id>')


class ControllerAvailableBlockSchema(ma.ModelSchema):
    """
    Available Blocks (not installed) living on a Controller
    """
    class Meta:
        model = ControllerBlock
        fields = ('type', 'object_id', 'name', 'last_seen')

    type = ma.Function(lambda obj: obj.__class__.__name__)
    last_seen = ma.Field(attribute='updated_at')



from datetime import datetime


class InfluxMeasurementWriter:
    """
    Utilitary class to ease writing measures to Influx
    """
    def __init__(self, influx_client, measurement):
        self.client = influx_client
        self.measurement = measurement

    def write(self, fields):
        json_body = self._to_json(fields)
        self.client.write_points(json_body)
        return True

    def _to_json(self, fields):
        json_body = [
            {
                "measurement": self.measurement,
                "time": datetime.utcnow(),
                "fields": fields
            }
        ]

        return json_body



from sqlalchemy import (
    Column, Integer, String, Boolean, ForeignKey,
    UniqueConstraint, DateTime
)
from zope.interface import implementer

from ..models import ControllerBlock

from .interfaces import ISwitchActuator


@implementer(ISwitchActuator)
class DigitalPin(ControllerBlock):
    __tablename__ = 'controller_block_digital_pin'

    __mapper_args__ = {
        'polymorphic_identity': "controller_block_digital_pin"
    }

    digital_pin_id = Column(Integer, ForeignKey('controller_block.id', ondelete="CASCADE"), primary_key=True)

    is_inverted = Column(Boolean)
    pin_number = Integer()

    def __repr__(self):
        return "<DigitalPin '{0}'>".format(self.name)



import logging

from brewpi_service.datasync.abstract import AbstractDeviceSyncher
from brewpi_service.database import db_session, get_or_create

from .models import ClockDevice

LOGGER = logging.getLogger("rq.worker")


class ScaledTimeSyncher(AbstractDeviceSyncher):
    """
    Handle synching with controller for device ClockDevice
    """
    def update(self, controller, state, event):
        LOGGER.debug(controller.system_id().read())
        device_id = int.from_bytes(event.idchain, byteorder="little", signed=False)

        # get or create model
        device, created = get_or_create(db_session, ClockDevice,
                                        create_method_kwargs={'time': state.time,
                                                              '_scale': state.scale},
                                        device_id=device_id,
                                        controller_id=1)

        if created is False:
            db_session.add(device)
            device.time = state.time
            device._scale = state.scale
            db_session.commit()
        else:
            LOGGER.info("Created <ClockDevice>(ID={0})".format(device_id))



import logging
import time

# from brewpi.controlbox.codecs.time import (
#    BrewpiStateCodec, BrewpiConstructorCodec
#)
# from brewpi.protocol.factory import all_sniffers
# from controlbox.connector_facade import ControllerDiscoveryFacade
from controlbox.protocol.io import determine_line_protocol
from controlbox.controller import Controlbox
# from brewpi.connector.controlbox.objects import BrewpiController, BuiltInObjectTypes
# from controlbox.events import (
#    ControlboxEvents, ConnectorCodec, ConnectorEventVisitor
# )

# from controlbox.connector.base import (
#    ConnectorConnectedEvent, ConnectorDisconnectedEvent
# )
# from controlbox.protocol.controlbox import ControlboxProtocolV1
# from controlbox.connector.socketconn import TCPServerEndpoint

from .database import DatabaseControllerSyncher

LOGGER = logging.getLogger(__name__)

ConnectorEventVisitor = object # FIXME

class BrewpiEvents(ConnectorEventVisitor):
    """
    Handle events for one controller
    """
    handlers = {}

    def __init__(self, aController):
        self._controller = aController

    def __call__(self, event):
        LOGGER.debug(event)
        if event.type == 1:  # XXX Should be a constant here, not int
            state_type = type(event.state)
            if state_type in self.handlers:
                self.handlers[state_type].update(self._controller,
                                                 event.state,
                                                 event)


def sniffer(conduit):
    return determine_line_protocol(conduit, all_sniffers)


class ControlboxSyncher(DatabaseControllerSyncher):
    """
    A loop that syncs to the controller using Controlbox/Connector
    """
    # discoveries = [
    #    ControllerDiscoveryFacade.build_serial_discovery(sniffer),
    #    ControllerDiscoveryFacade.build_tcp_server_discovery(sniffer, "brewpi",
    #                                                         known_addresses=()),
    # ]

    def __init__(self):
        self.facade = ControllerDiscoveryFacade(self.discoveries)

        self.facade.manager.events.add(self._handle_connection_event)  # connected?

    def _dump_device_info_events(self, connector, protocol):
        """
        XXX Document that
        """
        if not hasattr(protocol, 'controller'):
            protocol.controller = Controlbox(connector)
            events = protocol.controller.events = ControlboxEvents(protocol.controller, BrewpiConstructorCodec(), BrewpiStateCodec())
            controller = BrewpiController(connector, BuiltInObjectTypes())
            controller.initialize(True)
            events.listeners.add(BrewpiEvents(controller))
        else:
            events = protocol.controller.events

        # todo - would be good if the system itself described the type as part of the read response
        # this makes the system more self-describing
        events.read_system([0], 0)
        events.read_system([1], 1)

    def _handle_connection(self, connection):
        """
        XXX Doc
        """
        connector = connection.connector
        if connector.connected:
            self._dump_device_info_events(connector, connector.protocol)

    def _handle_connection_event(self, event):
        """
        Callback when a controller (dis)appears
        """
        if type(event) == ConnectorConnectedEvent:
            self.on_controller_appeared(event)
        elif type(event) == ConnectorDisconnectedEvent:
            self.on_controller_disappeared(event)

    def run(self):
        while True:
            self.facade.update()

            for connection in self.facade.manager.connections.values():
                self._handle_connection(connection)

            time.sleep(1)



from flask import jsonify

from brewpi_service.rest import (
    api_v1,
    marshal_with, MethodResource
)

from .models import TemperatureSensor, PID
from .schemas import (
    TemperatureSensorSchema,
    PIDSchema,
)

# -- Temperature Sensor
@marshal_with(TemperatureSensorSchema(many=True))
class TemperatureSensorList(MethodResource):
    """
    List Temperature Sensors
    """
    def get(self, **kwargs):
        return TemperatureSensor.query.all()

api_v1.register('/temperature_sensors/', TemperatureSensorList)


@marshal_with(TemperatureSensorSchema)
class TemperatureSensorDetail(MethodResource):
    """
    Detail a given Temperature Sensor
    """
    def get(self, id, **kwargs):
        return TemperatureSensor.query.get(id)

api_v1.register('/temperature_sensors/<id>/', TemperatureSensorDetail)


# -- PID
@marshal_with(PIDSchema(many=True))
class PIDList(MethodResource):
    """
    List PIDs
    """
    def get(self, **kwargs):
        return PID.query.all()

api_v1.register('/pids/', PIDList)


@marshal_with(PIDSchema)
class PIDDetail(MethodResource):
    """
    Detail a given PID
    """
    def get(self, id, **kwargs):
        return PID.query.get(id)

api_v1.register('/pids/<id>/', PIDDetail)



from zope.interface import Interface, Attribute

from ..interfaces import IBlock


class ISwitchActuator(IBlock):
    """
    Anything that can switch a state
    """
    is_inverted = Attribute("Is this pin inverted?")



import logging
from itertools import groupby
import time


from circuits import Component, handler

from brewpi_service.database import db_session
from brewpi_service.controller.models import Controller
from brewpi_service.controller.events import (
    ControllerConnected,
    ControllerBlockList
)

from brewpi_service.plugins_dir.temperature_sensor.models import TemperatureSensor, PID, SensorSetpointPair

from ..abstract import AbstractControllerSyncherBackend

from brewpi_service.controller.profile import ControllerProfileManager, NoResultFound
from brewpi_service.controller.state import ControllerStateManager

from brewpiv2.messages.temperature import TemperaturesMessage
from brewpiv2.messages.control import ControlSettingsMessage

from .brewpi_legacy import BrewPiLegacySyncherBackend

LOGGER = logging.getLogger(__name__)


class VirtualBrewPiSyncherBackend(Component, AbstractControllerSyncherBackend):
    """
    A virtual BrewPi Controller for testing purpose
    """
    def __init__(self, aControllerStateManager):
        super(VirtualBrewPiSyncherBackend, self).__init__()

        self.controller = None
        self.profile = None
        self.controller_state = None

        self._controller_state_manager = aControllerStateManager

        self.shutdown = False

    def make_profile(self, name="virtual_brewpi"):
        """
        Create the virtual profile
        """
        return BrewPiLegacySyncherBackend.make_profile(name)


    def started(self, *args):
        profile_name = "virtual_brewpi"
        try:
            self.profile = ControllerProfileManager.get(profile_name)
        except NoResultFound:
            self.profile = self.make_profile(name=profile_name)

        self.controller = Controller(name="Virtual BrewPi",
                                     uri="virtual://localhost",
                                     description="A virtual controller",
                                     profile=self.profile)

        self.controller_state = self._controller_state_manager.get_for_controller(self.controller)

        self.fire(ControllerConnected(self.controller))

        while not self.shutdown:
            # Send a list of available devices every second


            # If we get a message from the controller
            control_settings = ControlSettingsMessage(temp_format='C',
                                                      heater1_kp=0.8, heater1_ti=200, heater1_td=10,
                                                      heater1_infilter=12, heater1_dfilter=30, heater1_pwm_period=2,
                                                      heater2_kp=1.2, heater2_ti=100, heater2_td=3,
                                                      heater2_infilter=24, heater2_dfilter=10, heater2_pwm_period=1,
                                                      cooler_kp=2.0, cooler_ti=150, cooler_td=20,
                                                      cooler_infilter=10, cooler_dfilter=20, cooler_pwm_period=30,
                                                      min_cool_time=10, min_cool_idle_time=30,
                                                      beer2fridge_kp=0.1, beer2fridge_ti=2, beer2fridge_td=30,
                                                      beer2fridge_infilter=13, beer2fridge_dfilter=39, beer2fridge_pid_max=60,
                                                      deadtime=10)
            self.on_receive_control_settings(control_settings)


            # If we get a change from the service/user
            heater1pid = PID.query.filter(PID.name=="heater1pid").one()

            heater1pid.kp = 20

            transaction = self.controller_state.begin_transaction()
            transaction.add(heater1pid)
            self.controller_state.commit(self.controller,
                                         transaction) # fire message


            # Try to link a new sensor to heater2
            heater2pid = PID.query.filter(PID.name=="heater2pid").one()
            heater2pid.input = TemperatureSensor(object_id=99)

            transaction = self.controller_state.begin_transaction()
            transaction.add(heater2pid)
            self.controller_state.commit(self.controller,
                                         transaction) # fire message

            self.shutdown = True
            time.sleep(1)

    @handler("ControllerStateChangeRequest")
    def on_controller_state_change_request(self, event):
        for block in groupby(event.changes, key=lambda x: x[0]):

            control_changes = {}
            for change in block[1]:
                (block, fieldname, requested_value) = change
                if type(block) == PID:
                    ## SPECIFIC
                    if fieldname in ('input', ): # We have a link change
                        print("update link '{1}' on block {0} with object {2}".format(block, fieldname, requested_value))
                        if block.name in ('heater1pid', 'cooler1pid'):
                            InstallDeviceCommand(0, DeviceAssignation.CHAMBER, DeviceFunction.CHAMBER_TEMP, HardwareType.TEMP_SENSOR)
                        elif block.name == 'beer2fridgepid':
                            InstallDeviceCommand(0, DeviceAssignation.BEER, DeviceFunction.BEER_TEMP, HardwareType.TEMP_SENSOR)
                    elif fieldname in ('kp',):
                        print("collecting new data for {0} {1} {2}".format(block, fieldname, requested_value))
                        control_changes[fieldname] = requested_value
            if len(control_changes) > 0:
                print("sending control settings...")
                ControlSettingsCommand(heater1_kp=control_changes['kp'])

    def on_receive_control_settings(self, control_settings):
        """
        Feed the controller state system models from our backend values
        """
        try:
            heater1pid = self.profile.get_block_by_name('heater1pid')
        except NoResultFound:
            LOGGER.error("Profile mismatch on controller {0}".format(self.controller))

            return False

        self.controller_state.set(heater1pid, 'kp', control_settings.heater1_kp)

        return True

    # Event handlers
    @handler("ControllerRequestCurrentProfile")
    def on_request_current_profile(self):
        LOGGER.debug("Requesting current profile for controller {0}".format(self.controller))

        #self.fire(ControllerBlockList(self.controller,
        #                             [self.heater1_setpoint, self.temperature_sensor, self.heater1_pid]))



from collections import defaultdict
from datetime import datetime, timedelta


class AvailableControllerBlocksCache:
    """
    Cache the available blocks, in-memory.
    """
    def __init__(self):
        self._blocks = defaultdict(dict)

    def _hash_for(self, aControllerBlock):
        return hash((aControllerBlock.object_id,))

    def add(self, aController, aControllerBlock):
        block_hash = self._hash_for(aControllerBlock)

        if block_hash not in self._blocks[aController.uri].keys():
            self._blocks[aController.uri][block_hash] = aControllerBlock

        # If we already have this block in memory, mark it fresh
        self._blocks[aController.uri][block_hash].updated_at = datetime.utcnow()

    def remove(self, aController, aControllerBlock):
        block_hash = self._hash_for(aControllerBlock)

        if block_hash in self._blocks[aController.uri].keys():
            del self._blocks[aController.uri][block_hash]
            return True

        return False

    def get_all_for(self, aController):
        return self._blocks[aController.uri].values()

    def cleanup_stale_blocks_for(self, aController, timeout=timedelta(seconds=10)):
        """
        Remove all blocks that are outdated
        """
        now = datetime.utcnow()

        blocks_to_remove = []

        for block in self._blocks[aController.uri].values():
            if (now - block.updated_at) >= timeout:
                blocks_to_remove.append(block)

        for block in blocks_to_remove:
            self.remove(aController, block)




available_blocks_cache = AvailableControllerBlocksCache()



import logging
import platform
import time
from datetime import datetime, timedelta
from collections import OrderedDict

from circuits import handler, Component

from brewpiv2.commands import (
    ControlSettingsCommand,
    ListInstalledDevicesCommand,
    ListAvailableDevicesCommand
)
from brewpiv2.constants import (
    HardwareType
)
from brewpiv2.controller import (
    BrewPiControllerManager,
    BrewPiController,
    ControllerObserver,
    MessageHandler,
    message_received
)

from brewpi_service.database import db_session

from brewpiv2.messages.decoder import RawMessageDecoder

from brewpi_service.controller.profile import ControllerProfileManager, NoResultFound

from brewpi_service.controller.models import Controller
from brewpi_service.controller.events import (
    ControllerConnected,
    ControllerDisconnected,
    ControllerBlockList,
    ControllerCleanStaleAvailableBlocks
)
from brewpi_service.controller.profile import (
    ControllerProfileManager,
    NoResultFound
)
from brewpi_service.controller.state import ControllerStateManager
from brewpi_service.controller.blocks.models import DigitalPin

from brewpi_service.plugins_dir.temperature_sensor.models import TemperatureSensor, PID, SensorSetpointPair, SetpointSimple


from ..abstract import AbstractControllerSyncherBackend

from .cache import available_blocks_cache


LOGGER = logging.getLogger(__name__)


class AvailableDevicePool:
    """
    Patch Object to assign an ID to available devices (which otherwise defaults
    to -1). The goal is to help trace them when updating.
    """
    def __init__(self):
        self._devices = {}

    def find_free_id(self):
        used_ids = self._devices.values()
        for i in range(10, 100):
            if -i not in used_ids:
                return -i

        raise Exception # FIXME

    def _hash_device(self, aDevice):
        return hash((aDevice.slot, aDevice.hardware_type, aDevice.pin, aDevice.address))

    def get_id_for(self, aDevice):
        id = None
        try:
            id = self._devices[self._hash_device(aDevice)]
        except KeyError:
            new_id = self.find_free_id()
            self._devices[self._hash_device(aDevice)] = new_id
            id = new_id

        return id


class BrewPiLegacySyncherBackend(Component, AbstractControllerSyncherBackend):
    """
    A loop that syncs a BrewPi Controller using the legacy firmware
    """
    def __init__(self, aControllerStateManager):
        super(BrewPiLegacySyncherBackend, self).__init__()

        self.manager = BrewPiControllerManager()
        self.msg_decoder = RawMessageDecoder()

        self._controller_state_manager = aControllerStateManager

        self.controller_observers = []

        self.shutdown = False

    @staticmethod
    def make_profile(name="brewpi_legacy"):
        """
        Create the brewpiv2 profile
        """
        LOGGER.info("Making BrewPi Legacy Profile into database under name '{0}'...".format(name))
        profile = ControllerProfileManager.create(name)

        beer1_setpoint = SetpointSimple(profile=profile,
                                        is_static=True,
                                        object_id=51,
                                        name="beer1set")
        db_session.add(beer1_setpoint)

        beer1_setpoint_pair = SensorSetpointPair(profile=profile,
                                                 name="heater1sensorsetpointpair",
                                                 setpoint=beer1_setpoint,
                                                 is_static=True,
                                                 # sensor=beer2_sensor,
                                                 object_id=61) # not required since static

        db_session.add(beer1_setpoint_pair)


        beer2_setpoint = SetpointSimple(profile=profile,
                                        is_static=True,
                                        object_id=52,
                                        name="beer2set")
        db_session.add(beer2_setpoint)

        beer2_setpoint_pair = SensorSetpointPair(profile=profile,
                                                 name="heater2sensorsetpointpair",
                                                 setpoint=beer2_setpoint,
                                                 is_static=True,
                                                 # sensor=beer2_sensor,
                                                 object_id=62) # not required since static

        db_session.add(beer2_setpoint_pair)

        # heater2_pwm = ActuatorPwm(profile=profile,
        #                           object_id=53,
        #                           period=4,
        #                           name="heater2pwm")

        # db_session.add(heater2_pwm)

        heater1_pid = PID(profile=profile,
                          name="heater1pid",
                          object_id=71,
                          is_static=True,
                          input=beer1_setpoint_pair)
                          # output=heater2_pwm)


        heater2_pid = PID(profile=profile,
                          name="heater2pid",
                          object_id=72,
                          is_static=True,
                          input=beer2_setpoint_pair)
                          # output=heater2_pwm)

        db_session.add(heater1_pid)
        db_session.add(heater2_pid)

        db_session.commit()

        return profile


    @handler("stopped")
    def stopped(self, *args):
        print("STOPPPEED")
        self.shutdown = True

    @handler("init")
    def init(self, *args):
        # Make sure we have our profile
        profile_name = "brewpi_legacy"
        try:
            self.profile = ControllerProfileManager.get(profile_name)
        except NoResultFound:
            self.profile = self.make_profile(name=profile_name)

        # wifi_ctrl = BrewPiController("socket://192.168.0.46:6666")
        # self.manager.controllers[wifi_ctrl.serial_port] = wifi_ctrl
        # controller_observer = BrewPiLegacyControllerObserver(wifi_ctrl,
        #                                                      self.profile,
        #                                                      self._controller_state_manager).register(self)
        # wifi_ctrl.subscribe(controller_observer)
        # wifi_ctrl.connect()

        # Remove all available devices

    @handler("started")
    def started(self, *args):
        last_update = datetime.utcnow()

        while not self.shutdown:
            if (datetime.utcnow() - last_update) > timedelta(seconds=5):
                LOGGER.debug("Requesting controller update from backend {0}".format(self))

                # Request list of available devices with values
                for port, controller in self.manager.controllers.items():
                    if controller.is_connected:
                        controller.send(ListAvailableDevicesCommand(with_values=True))
                        controller.send(ListInstalledDevicesCommand(with_values=True))

                # Clean up every cycle
                for observer in self.controller_observers:
                    observer.cleanup(last_update)

                last_update = datetime.utcnow()

            # Update manager
            for new_controller in self.manager.update():
                controller_observer = BrewPiLegacyControllerObserver(new_controller,
                                                                     self.profile,
                                                                     self._controller_state_manager).register(self)

                self.controller_observers.append(controller_observer)

                new_controller.subscribe(controller_observer)
                new_controller.connect()

            # Process messages from controllers
            for port, controller in self.manager.controllers.items():
                if controller.is_connected:
                    for raw_message in controller.process_messages():
                        for msg in self.msg_decoder.decode_controller_message(raw_message):
                            controller.notify(message_received, msg)

            # Dispatch updates to backstores
            for observer in self.controller_observers:
                observer.dispatch_and_clear_updates()

            yield

    def __str__(self):
        return "BrewPi Legacy Backend"


class SyncherMessageHandler(MessageHandler):
    """
    Take actions upon Controller Message reception
    """
    def __init__(self, aController, aControllerProfile):
        self.updated_blocks = []
        self.controller_profile = aControllerProfile
        self.controller = aController

        self.available_device_pool = AvailableDevicePool()

    def clear_updates(self):
        """
        Once updates have been treated, clear them for next update
        """
        self.updated_blocks = []

    def installed_device(self, aDevice):
        if aDevice.hardware_type == HardwareType.TEMP_SENSOR:
            self.updated_blocks.append(TemperatureSensor(object_id=aDevice.slot,
                                                         profile=self.controller_profile,
                                                         profile_id=self.controller_profile.id,
                                                         name="1-Wire Temperature Sensor@{0}".format(aDevice.address),
                                                         value=(aDevice.value, aDevice.value)))
        else:
            LOGGER.warn("Unknown installed device, ignoring.")

    def available_device(self, aDevice):
        if aDevice.hardware_type == HardwareType.DIGITAL_PIN:
            available_blocks_cache.add(self.controller,
                                       DigitalPin(object_id=self.available_device_pool.get_id_for(aDevice),
                                                  name="Digital Pin {0}".format(aDevice.pin),
                                                  pin_number=aDevice.pin,
                                                  is_inverted=aDevice.pin_inverted))

        elif aDevice.hardware_type == HardwareType.TEMP_SENSOR:
            available_blocks_cache.add(self.controller,
                                       TemperatureSensor(object_id=self.available_device_pool.get_id_for(aDevice),
                                                         value=(aDevice.value, aDevice.value), # Set actual value
                                                         name="1-Wire Temperature Sensor@{0}".format(aDevice.address)))
        else:
            LOGGER.debug("Unknown device to synch: {0}".format(aDevice))

    def uninstalled_device(self, anUninstalledDeviceMessage):
        LOGGER.warn("TBI: update uninstalled device!")

    def log_message(self, aLogMessage):
        LOGGER.warn("TBI: Log message !")

    def control_settings(self, aControlSettingsMessage):
        LOGGER.warn("TBI: Control Settings!")

    def control_constants(self, aControlConstantsMessage):
        LOGGER.warn("TBI: Control Constants!")

    def temperatures(self, aTemperaturesMessage):
        LOGGER.warn("TBI: Temperatures!")


class BrewPiLegacyControllerObserver(Component, ControllerObserver):
    """
    Controller event handler for the Legacy backend.

    Receives events from the backend and dispatch them as events the service
    can understand.
    """
    def __init__(self, aController, aControllerProfile, aControllerStateManager):
        super(BrewPiLegacyControllerObserver, self).__init__()

        self.controller_device = aController
        self.controller = None
        self.profile = aControllerProfile
        self.controller_state_manager = aControllerStateManager
        self.controller_state = None

        self.msg_handler = None

    def _make_controller_uri(self, aBrewPiController):
        """
        Forge the controller uri as a string
        """
        return "{0}:{1}".format(platform.node(),
                                aBrewPiController.serial_port)

    @handler("ControllerStateChangeRequest")
    def on_controller_state_change_request(self, event):
        for change in event.changes:
            (block, fieldname, requested_value) = change
            if type(block) == PID:
                ## FIXME SPECIFIC
                print("post changes to controller {0} with {1} {2} {3}".format(self.controller, block, fieldname, requested_value))
                self.controller_device.send(ControlSettingsCommand(heater1_kp=requested_value))

    def _on_message_received(self, aMessage):
        """
        When a message is received from the controller
        """
        self.msg_handler.accept(aMessage)

    def dispatch_and_clear_updates(self):
        if not self.msg_handler:
            return

        if len(self.msg_handler.updated_blocks) > 0:
            self.fire(ControllerBlockList(self.controller, self.msg_handler.updated_blocks))

        self.msg_handler.clear_updates()

    def _on_controller_connected(self, aBrewPiController):
        controller = Controller(name="BrewPi with legacy firmware on {0} at {1}".format(platform.node(),
                                                                                        aBrewPiController.serial_port),
                                profile_id=self.profile.id,
                                uri=self._make_controller_uri(aBrewPiController),
                                description="A BrewPi connected to a serial port, using the legacy protocol.",
                                connected=aBrewPiController.is_connected)

        self.controller = controller

        self.fire(ControllerConnected(controller))

        self.controller_state = self.controller_state_manager.get_for_controller(self.controller)

        # Create a message handler for this controller
        self.msg_handler = SyncherMessageHandler(self.controller, self.profile)

        # If we get a change from the service/user
        # heater2pid = PID.query.filter(PID.profile==self.profile,
        #                               PID.name=="heater2pid").one()

        # heater2pid.kp = 20

        # transaction = self.controller_state.begin_transaction()
        # transaction.add(heater2pid)
        # self.controller_state.commit(self.controller,
        #                              transaction) # fire message

    def _on_controller_disconnected(self, aBrewPiController):
        """
        Callback when a controller has been disconnected
        """
        self.fire(ControllerDisconnected(self.controller))

    def cleanup(self, limit_time):
        if self.controller:
            available_blocks_cache.cleanup_stale_blocks_for(self.controller)
            self.fire(ControllerCleanStaleAvailableBlocks(self.controller, limit_time))



from unittest.mock import Mock

from brewpi_service.datasync.backends.brewpi_legacy import AvailableDevicePool

class TestAvailableDevicePool:
    def test_simple_insert(self):
        pool = AvailableDevicePool()

        device1 = Mock()

        id = pool.get_id_for(device1)

        assert(type(id) is int)

    def test_same_insert(self):
        pool = AvailableDevicePool()

        device1 = Mock()

        id1 = pool.get_id_for(device1)
        id2 = pool.get_id_for(device1)

        assert(id1 == id2)

    def test_two_inserts(self):
        pool = AvailableDevicePool()

        device1 = Mock()
        device2 = Mock()

        id1 = pool.get_id_for(device1)
        id2 = pool.get_id_for(device2)

        assert(id1 != id2)




import functools

import marshmallow

from flask_apispec import (
    FlaskApiSpec,
    marshal_with,
    MethodResource
)


class Api:
    """
    Holds a version of the API with documentation and allows routes to
    register.
    """
    def __init__(self, prefix=""):
        self.prefix = prefix
        self._deferred_routes = []

    def register(self, path, aMethodResource):
        self._deferred_routes.append(
            ("/{0}/{1}".format(self.prefix, path.lstrip('/')),
             aMethodResource.as_view(name=aMethodResource.__name__.lower())
             )
        )

    def init_app(self, app):
        for deferred in self._deferred_routes:
            app.add_url_rule(deferred[0], view_func=deferred[1])
            specs.register(deferred[1])


specs = FlaskApiSpec()
api_v1 = Api(prefix="v0.1")



import logging

from circuits import Component, handler

from sqlalchemy.exc import IntegrityError

from sqlalchemy.orm import scoped_session, sessionmaker

from brewpi_service.database import engine, get_or_create, db_session
from brewpi_service.controller.models import Controller
from brewpi_service.controller.models import ControllerBlock

from ..abstract import AbstractBackstoreSyncher


LOGGER = logging.getLogger(__name__)


class DatabaseSyncher(Component, AbstractBackstoreSyncher):
    """
    Synchronize backend events to a Database using SQLAlchemy
    """

    def init(self):
        self.session = scoped_session(sessionmaker(autocommit=False,
                                                   autoflush=False,
                                                   bind=engine))

    @handler("started")
    def started(self, *args):
        # Mark all controllers as disconnected
        self.session.query(Controller).update({Controller.connected: False})
        self.session.commit()

    @handler("ControllerConnected")
    def on_controller_appeared(self, event):
        aController = event.controller

        controller, created = get_or_create(self.session, Controller,
                                            create_method_kwargs={'name': aController.name,
                                                                  'description': aController.description,
                                                                  'connected': True,
                                                                  'profile_id': aController.profile_id},
                                            uri=aController.uri)

        if created is False:
            LOGGER.debug("Controller has reconnected: {0}".format(controller))
            controller.connected = True

            self._wipe_available_blocks_for(controller.profile, self.session)

        else:
            LOGGER.debug("New Controller connected: {0}".format(controller))


        self.session.commit()

    def _wipe_available_blocks_for(self, aControllerProfile, aDBSession):
        """
        Remove all available blocks
        """
        aDBSession.query(ControllerBlock).filter(ControllerBlock.profile_id==aControllerProfile.id,
                                                 ControllerBlock.is_static==False).delete(synchronize_session=False)

    def _clean_stale_available_blocks_for(self, aControllerProfile, time_limit, aDBSession):
        """
        Remove all stale available blocks (not updated since a few seconds)
        """
        aDBSession.query(ControllerBlock).filter(ControllerBlock.profile_id==aControllerProfile.id,
                                                 ControllerBlock.is_static==False,
                                                 ControllerBlock.updated_at<=time_limit).delete(synchronize_session=False)

    @handler("ControllerDisconnected")
    def on_controller_disappeared(self, event):
        aController = event.controller

        LOGGER.debug("Controller disconnected: {0}".format(aController.uri))
        controller = db_session.query(Controller).filter(Controller.uri==aController.uri).first()
        controller.connected = False

        self._wipe_available_blocks_for(controller.profile, db_session)

        db_session.commit()

    @handler("ControllerCleanStaleAvailableBlocks")
    def on_clean_stale_available_blocks(self, event):
        LOGGER.debug("Cleaning Stale Available Blocks for controller {0}".format(event.controller))

        controller = db_session.query(Controller).filter(Controller.uri==event.controller.uri).first()
        self._clean_stale_available_blocks_for(controller.profile, event.time_limit, db_session)
        db_session.flush()

    @handler("ControllerBlockList") ## Disabled while working on available blocks
    def on_controller_block_list(self, event):
        LOGGER.debug("Synching {0} blocks from controller {1}".format(len(event.blocks), event.controller))

        db_controller = db_session.query(Controller).filter(Controller.uri==event.controller.uri).first()

        if db_controller is None:
            LOGGER.error("Got list of blocks for an unknown controller, aborting.")
            return # FIXME

        for block in event.blocks:
            # Check if we already have a block of this ID
            existing_block = db_session.query(ControllerBlock).filter(ControllerBlock.profile_id==block.profile.id,
                                                                      ControllerBlock.object_id==block.object_id).first()

            if existing_block is not None:
                # Check we have the same type in DB as we have in controller's memory
                typed_block = db_session.query(type(block)).filter(ControllerBlock.profile_id==block.profile.id,
                                                                   ControllerBlock.object_id==block.object_id).one()

                if typed_block is None:
                    # FIXME: Type mismatch, what should we do?
                    LOGGER.error("Type mismatch between Controller and Database, fix that by hand!")

                # just mark it updated
                from datetime import datetime
                existing_block.updated_at = datetime.utcnow()
                db_session.add(existing_block)
            else:
                LOGGER.debug("Adding block :{0}".format(block))
                db_session.add(block)

                LOGGER.debug("New object synched from controller memory: {0}".format(block))

        try:
            db_session.commit()
        except IntegrityError as e:
            db_session.rollback()
            LOGGER.error(e)




import logging

from influxdb import InfluxDBClient, SeriesHelper

from brewpi_service.plugins.core import BrewPiServicePlugin
from brewpi_service.database import db_session, get_or_create
from brewpi_service.admin import admin, ModelView

from .models import LoggedDeviceConfiguration
from .utils import InfluxMeasurementWriter

__plugin__ = "InfluxLoggerPlugin"

LOGGER = logging.getLogger(__name__)


class InfluxLoggerPlugin(BrewPiServicePlugin):
    """
    Log any model property to influxdb time series
    """
    def __init__(self, *args, **kwargs):
        self.client = None

        super(InfluxLoggerPlugin, self).__init__(*args, **kwargs)

    def log_data_before_update(self, mapper, connection, target):
        """
        Callback when an object is about to be saved: log its data if a
        `LoggedDeviceConfiguration` matches.
        """
        logging_configuration = db_session.query(LoggedDeviceConfiguration).filter(LoggedDeviceConfiguration.device_id==target.id).one()
        if not logging_configuration:
            return

        writer = InfluxMeasurementWriter(self.client, measurement='object_{0}'.format(logging_configuration.device_id))

        # XXX Should support multiple fields
        field = logging_configuration.device_field
        writer.write({field: getattr(target, field)})

    def setup(self):
        from .models import LoggedDeviceConfiguration

    def install(self):
        # Admin
        from .admin import LoggedDeviceConfigurationAdminView

        self.client = InfluxDBClient('localhost', 8086, 'root', 'root', 'brewpi')
        self.client.create_database('brewpi')
        self.client.switch_database('brewpi')

        from .models import LoggedDeviceConfiguration

        from sqlalchemy import event

        # for configuration in db_session.query(LoggedDeviceConfiguration):
        #    event.listen(configuration.device.__class__, 'before_update', self.log_data_before_update)




from sqlalchemy import (
    Column, Integer,
    String, ForeignKey
)
from sqlalchemy.orm import relationship, backref

from brewpi_service.database import Base
from brewpi_service.controller.models import ControllerBlock


class LoggedDeviceConfiguration(Base):
    """
    Configurable Watched Devices so they are logged into influxdb
    """
    __tablename__ = 'influx_logged_devices'

    id = Column(Integer, primary_key=True)

    block_id = Column(Integer, ForeignKey('controller_block.id'), nullable=False)
    block = relationship(ControllerBlock,
                         backref=backref('logging_configurations',
                                         uselist=True,
                                         cascade='delete, all'))

    device_field = Column(String, nullable=False)

    def __repr__(self):
        return '<Device logging configuration for {0})>'.format(self.device)



from brewpi_service import ma
from brewpi_service.controller.schemas import ControllerBlockSchema

from .models import ClockDevice


class ClockSchema(ControllerBlockSchema):
    """
    A clock that ticks
    """
    class Meta:
        model = ClockDevice
        fields = ('id', 'scale', 'time', 'url')

    scale = ma.Integer(attribute='_scale')
    url = ma.AbsoluteUrlFor('clockdevice.details_view', id='<id>')



from setuptools import setup

setup(
    name='brewpi-service',
    version='0.1',
    long_description=__doc__,
    packages=['brewpi_service'],
    include_package_data=True,
    zip_safe=False,
    install_requires=['Flask']
)



from brewpi_service import ma

from zope.interface import implementer

from ..schemas import ControllerBlockDisambiguator, ControllerBlockSchema

from .interfaces import ISwitchActuator

from .models import DigitalPin


@implementer(ISwitchActuator)
class DigitalPinSchema(ControllerBlockSchema):
    """
    Serialization schema for a Digital Pin
    """
    class Meta:
        model = DigitalPin
        fields = ControllerBlockSchema.Meta.fields


ControllerBlockDisambiguator.class_to_schema[DigitalPin.__name__] = DigitalPinSchema



from collections import defaultdict, UserList

from circuits import Component

from .events import ControllerStateChangeRequest

import logging

LOGGER = logging.getLogger(__name__)


class ControllerDataField(UserList):
    (ACTUAL, REQUESTED, DIRTY) = range(3)

    def __init__(self, *args, **kwargs):
        self._writable = False

        if 'writable' in kwargs:
            self._writable = True
            kwargs.pop('writable')

        self._args = args
        self._kwargs = kwargs

        super(ControllerDataField, self).__init__([None, None, False])

    def set_actual_value(self, value):
        self.data[ControllerDataField.ACTUAL] = value

    def request_value(self, value):
        if value != self.get_actual_value():
            self.data[ControllerDataField.REQUESTED] = value
            self.data[ControllerDataField.DIRTY] = True

    def get_requested_value(self):
        return self.data[ControllerDataField.REQUESTED]

    def get_actual_value(self):
        return self.data[ControllerDataField.ACTUAL]


class BlockState(defaultdict):
    def __init__(self):
        super(BlockState, self).__init__(lambda: [None, None])


class ControllerState(Component):
    """
    The simplest, most universal way of representing a controller state: a
    dictionary of block IDs themselfves composed of dictionaries of attributes.
    Example:
      - {BLOCK_ID: {ATTR_NAME: (CURRENT_VALUE, REQUESTED_VALUE)}}
    """
    def __init__(self, aControllerStateManager):
        self.states = defaultdict(BlockState)

        self._controller_state_manager = aControllerStateManager

    def set(self, aBlock, attribute_name, value):
        self.states[aBlock.object_id][attribute_name] = [value, None]

    def set_requested(self, aBlock, attribute_name, value):
        attr_data = self.get(aBlock, attribute_name)
        attr_data[1] = value

    def get(self, aBlock, attribute_name):
        return self.states[aBlock.object_id][attribute_name]

    def get_requested(self, aBlock, attribute_name):
        return self.get(aBlock, attribute_name)[1]

    def get_actual(self, aBlock, attribute_name):
        return self.get(aBlock, attribute_name)[0]

    def begin_transaction(self):
        return ControllerStateTransaction()

    def commit(self, aController, aControllerStateTransaction):
        changes = aControllerStateTransaction.get_compiled_changes()

        self._controller_state_manager.commit_state(aController, changes)

        # Empty transaction
        aControllerStateTransaction.clear() # as callback?


class ControllerStateTransaction:
    def __init__(self):
        self.dirty_blocks = []

    def add(self, aBlock):
        self.dirty_blocks.append(aBlock)

    def clear(self):
        """
        Clear transaction
        """
        self.dirty_blocks = []

    def get_compiled_changes(self):
        # Collect changed fields
        changes = []

        for block in self.dirty_blocks:
            for fieldname in block.get_dirty_fields():
                field = getattr(block, fieldname)
                changes.append((block, fieldname, field.get_requested_value()))

        return changes

class ControllerStateManager(Component):
    """
    The manager holding states for every Controller registered
    Shoud be seen as a Singleton
    """
    def __init__(self):
        self.states = defaultdict(lambda: ControllerState(self))
        super(ControllerStateManager, self).__init__()

    def get_for_controller(self, aController):
        return self.states[aController]

    def commit_state(self, aController, changes):
        state = self.get_for_controller(aController)
        for change in changes:
            state.set_requested(change[0], change[1], change[2])

        LOGGER.debug("Commiting changes for controller {0}".format(aController))
        self.fire(ControllerStateChangeRequest(aController, changes))



import logging

from brewpi_service.plugins.core import BrewPiServicePlugin

__plugin__ = "TemperatureSensorPlugin"

LOGGER = logging.getLogger(__name__)


class TemperatureSensorPlugin(BrewPiServicePlugin):
    """
    A Plugin that adds everything needed for the DS2xxx temperature sensors to
    work with the service.
    """

    def setup(self):
        from .models import TemperatureSensor # noqa
        from .models import PID # noqa
        from .models import SensorSetpointPair # noqa

    def install(self):
        # Admin
        from . import admin

        # REST Api
        from .models import TemperatureSensor, PID, SetpointSimple, SensorSetpointPair
        from .schemas import TemperatureSensorSchema, PIDSchema, SetpointSimpleSchema, SensorSetpointPairSchema
        from brewpi_service.controller.schemas import ControllerBlockDisambiguator

        ControllerBlockDisambiguator.class_to_schema[SensorSetpointPair.__name__] = SensorSetpointPairSchema
        ControllerBlockDisambiguator.class_to_schema[TemperatureSensor.__name__] = TemperatureSensorSchema
        ControllerBlockDisambiguator.class_to_schema[PID.__name__] = PIDSchema
        ControllerBlockDisambiguator.class_to_schema[SetpointSimple.__name__] = SetpointSimpleSchema




from marshmallow import fields

class ControllerData(fields.Field):
    """
    Serialize a ControllerDataField that is made of three element.
    Only outputs the actual value.
    """
    def _serialize(self, field, attr, obj):
        return field.get_actual_value()
